from __future__ import annotations

import datetime
import enum
import warnings
import typing
from uuid import uuid4


__all__ = (
    "ValueObject",
    "BatchName",
    "Days",
    "DaysToKeep",
    "DbUri",
    "ExecutionMillis",
    "Failure",
    "Flag",
    "JobName",
    "LogLevel",
    "LogLevelOption",
    "LogMessage",
    "MaxProcesses",
    "MaxRetries",
    "MinSecondsBetweenRefreshes",
    "MinSecondsBetweenTests",
    "NonEmptyStr",
    "Password",
    "PositiveInt",
    "ResourceName",
    "Result",
    "SchemaName",
    "SecondsSinceLastRefresh",
    "SingleChar",
    "Success",
    "TestName",
    "TimeoutSeconds",
    "Timestamp",
    "UniqueId",
)


class ValueObject:
    __slots__ = ("value",)

    def __init__(self, value: typing.Any, /):
        self.value = value

    def __eq__(self, other: object) -> bool:
        if other.__class__ is self.__class__:
            return self.value == typing.cast(ValueObject, other).value
        else:
            return NotImplemented

    def __ne__(self, other: object) -> bool:
        result = self.__eq__(other)
        if result is NotImplemented:
            return NotImplemented
        else:
            return not result

    def __lt__(self, other: object) -> bool:
        if other.__class__ is self.__class__:
            return self.value < typing.cast(ValueObject, other).value
        else:
            return NotImplemented

    def __le__(self, other: object) -> bool:
        if other.__class__ is self.__class__:
            return self.value <= typing.cast(ValueObject, other).value
        else:
            return NotImplemented

    def __gt__(self, other: object) -> bool:
        if other.__class__ is self.__class__:
            return self.value > typing.cast(ValueObject, other).value
        else:
            return NotImplemented

    def __ge__(self, other: object) -> bool:
        if other.__class__ is self.__class__:
            return self.value >= typing.cast(ValueObject, other).value
        else:
            return NotImplemented

    def __hash__(self) -> int:
        return hash(self.value)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.value!r})"

    def __str__(self) -> str:
        return str(self.value)


class Flag(ValueObject):
    def __init__(self, value: bool, /):
        if value is None:
            raise ValueError(
                f"{self.__class__.__name__} value is required, but got None."
            )

        if not isinstance(value, bool):
            raise TypeError(
                f"{self.__class__.__name__} expects an integer, but got {value!r}"
            )

        super().__init__(value)


class OptionalPositiveInt(ValueObject):
    def __init__(self, value: typing.Optional[int], /):
        if value is not None:
            if isinstance(value, int):
                if value < 0:
                    raise ValueError(
                        f"{self.__class__.__name__} value must be positive, but got {value!r}."
                    )
            else:
                raise TypeError(
                    f"{self.__class__.__name__} expects an integer, but got {value!r}"
                )

        super().__init__(value)


class PositiveInt(ValueObject):
    def __init__(self, value: int, /):
        if value is None:
            raise ValueError(
                f"{self.__class__.__name__} value is required, but got None."
            )
        elif isinstance(value, int):
            if value < 0:
                raise ValueError(
                    f"{self.__class__.__name__} value must be positive, but got {value!r}."
                )
        else:
            raise TypeError(
                f"{self.__class__.__name__} expects an integer, but got {value!r}"
            )

        super().__init__(value)


class NonEmptyStr(ValueObject):
    def __init__(self, value: str, /):
        if value is None:
            raise ValueError(
                f"{self.__class__.__name__} value is required, but got None."
            )
        elif isinstance(value, str):
            if not value:
                raise ValueError(
                    f"{self.__class__.__name__} value is required, but got {value!r}."
                )
        else:
            raise TypeError(
                f"{self.__class__.__name__} expects a string, but got {value!r}."
            )

        super().__init__(value)


class Success(ValueObject):
    def __init__(self) -> None:
        super().__init__(None)

    def __repr__(self) -> str:
        return "Success()"

    def __str__(self) -> str:
        return "Success"


class Failure(ValueObject):
    def __init__(self, message: str, /) -> None:
        value = NonEmptyStr(message).value
        super().__init__(value)


class Result(ValueObject):
    def __init__(self, value: typing.Union[Failure, Success], /) -> None:
        if value is None:
            raise ValueError(
                f"{self.__class__.__name__} value is required, but got None."
            )

        if not isinstance(value, (Failure, Success)):
            raise TypeError(
                f"{self.__class__.__name__} expects either a Success or Failure instance, "
                f"but got {value!r}."
            )

        super().__init__(value)

    @classmethod
    def failure(cls, message: str, /) -> Result:
        return Result(Failure(message))

    @classmethod
    def success(cls) -> Result:
        return Result(Success())

    @property
    def failure_message(self) -> str:
        if self.is_failure:
            return self.value.value
        else:
            raise TypeError(
                f"{self.__class__.__name__} does not contain a failure value, so it does not have "
                f"a failure message.  The value of the Result is {self.value!r}."
            )

    @property
    def failure_message_or_none(self) -> typing.Optional[str]:
        if self.is_failure:
            return self.value.value
        else:
            return None

    @property
    def is_failure(self) -> bool:
        return isinstance(self.value, Failure)

    @property
    def is_success(self) -> bool:
        return not self.is_failure


class UniqueId(ValueObject):
    def __init__(self, value: str, /):
        if value is None:
            raise ValueError(
                f"{self.__class__.__name__} value is required, but got None."
            )
        elif isinstance(value, str):
            if len(value) != 32:
                raise ValueError(
                    f"{self.__class__.__name__} value must be 32 characters long, "
                    f"but got {value!r}."
                )
            if not value.isalnum():
                raise ValueError(
                    f"{self.__class__.__name__} value must be all alphanumeric characters, but "
                    f"got {value!r}."
                )
        else:
            raise TypeError(
                f"{self.__class__.__name__} expects a str, but got {value!r}"
            )

        super().__init__(value)

    @classmethod
    def generate(cls) -> UniqueId:
        return UniqueId(uuid4().hex)


class SchemaName(ValueObject):
    def __init__(self, value: typing.Optional[str], /):
        if value is None:
            ...
        elif isinstance(value, str):
            if not value:
                raise ValueError(
                    f"If a {self.__class__.__name__} value is provided, then it must be at least 1 "
                    f"character long, but got {value!r}."
                )
        else:
            raise TypeError(
                f"{self.__class__.__name__} expects a str, but got {value!r}"
            )

        super().__init__(value)


class SingleChar(ValueObject):
    def __init__(self, value: str, /):
        if not value:
            raise ValueError(
                f"{self.__class__.__name__} value is required, but got {value!r}."
            )
        elif isinstance(value, str):
            if len(value) != 1:
                raise ValueError(
                    f"{self.__class__.__name__} must be 1 char long, but got {value!r}."
                )
        else:
            raise TypeError(
                f"{self.__class__.__name__} expects a str, but got {value!r}"
            )

        super().__init__(value)


class _DbName(ValueObject):
    def __init__(self, value: str, /):
        if value is None:
            raise ValueError(
                f"{self.__class__.__name__} value is required, but got None."
            )
        elif isinstance(value, str):
            if len(value) < 3 or len(value) >= 200:
                raise ValueError(
                    f"{self.__class__.__name__} must be between 3 and 200 characters long, but got "
                    f"{value!r}."
                )
        else:
            raise TypeError(
                f"{self.__class__.__name__} expects a str, but got {value!r}"
            )

        super().__init__(value)


class DbUri(NonEmptyStr):
    def __init__(self, value: str, /):
        super().__init__(value)


class BatchName(_DbName):
    def __init__(self, value: str, /):
        super().__init__(value)


class JobName(_DbName):
    def __init__(self, value: str, /):
        super().__init__(value)


class ResourceName(NonEmptyStr):
    ...


class MinSecondsBetweenRefreshes(ValueObject):
    def __init__(self, value: int, /):
        if isinstance(value, int):
            if value < 0:
                raise ValueError(
                    f"If a {self.__class__.__name__} must be >= 0, but got {value!r}."
                )
        else:
            raise TypeError(
                f"{self.__class__.__name__} expects an int, but got {value!r}"
            )

        super().__init__(value)


class MinSecondsBetweenTests(ValueObject):
    def __init__(self, value: int, /):
        if isinstance(value, int):
            if value < 0:
                raise ValueError(
                    f"If a {self.__class__.__name__} must be >= 0, but got {value!r}."
                )
        else:
            raise TypeError(
                f"{self.__class__.__name__} expects an int, but got {value!r}"
            )

        super().__init__(value)


class TestName(ValueObject):
    def __init__(self, value: str, /):
        if value is None:
            raise ValueError(
                f"{self.__class__.__name__} value is required, but got None."
            )
        elif isinstance(value, str):
            if len(value) < 3 or len(value) > 200:
                raise ValueError(
                    f"{self.__class__.__name__} must be between 3 and 200 characters long, "
                    f"but got {value!r}."
                )
        else:
            raise TypeError(
                f"{self.__class__.__name__} expects a str, but got {value!r}"
            )

        super().__init__(value)


class Days(PositiveInt):
    ...


class DaysToKeep(PositiveInt):
    ...


class ExecutionMillis(PositiveInt):
    ...

    @staticmethod
    def calculate(*, start_time: Timestamp, end_time: Timestamp) -> ExecutionMillis:
        elapsed_millis = int((end_time.value - start_time.value).total_seconds() * 1000)
        return ExecutionMillis(elapsed_millis)


class TimeoutSeconds(ValueObject):
    def __init__(self, value: typing.Optional[int], /):
        if value is None:
            ...
        elif isinstance(value, int):
            if value < 0:
                raise ValueError(
                    f"If a value is provided for {self.__class__.__name__}, then it must be positive."
                )
        else:
            raise TypeError(
                f"{self.__class__.__name__} expects an int, but got {value!r}"
            )

        super().__init__(value)


class MaxProcesses(OptionalPositiveInt):
    ...


class MaxRetries(PositiveInt):
    ...


class Timestamp(ValueObject):
    def __init__(self, value: datetime.datetime, /):
        if value is None:
            raise ValueError(
                f"{self.__class__.__name__} value is required, but got None."
            )

        if not isinstance(value, datetime.datetime):
            raise TypeError(
                f"{self.__class__.__name__} expects a datetime.datetime, but got {value!r}"
            )

        super().__init__(value)

    @classmethod
    def now(cls) -> Timestamp:
        return Timestamp(datetime.datetime.now())


class Password(ValueObject):
    def __init__(self, value: str, /):
        if value is None:
            raise ValueError(
                f"{self.__class__.__name__} value is required, but got None."
            )
        elif isinstance(value, str):
            if not value.strip():
                raise ValueError(
                    f"{self.__class__.__name__} value is required, but got {value!r}."
                )
        else:
            raise TypeError(
                f"{self.__class__.__name__} expects a str, but got {value!r}"
            )

        super().__init__(value)

    def __repr__(self) -> str:
        return f"Password({'*' * 10})"

    def __str__(self) -> str:
        return "*" * 10


class LogLevelOption(enum.Enum):
    Debug = 1
    Info = 2
    Error = 3


class LogLevel(ValueObject):
    def __init__(self, value: LogLevelOption, /):
        if value is None:
            raise ValueError(
                f"{self.__class__.__name__} value is required, but got None."
            )

        if not isinstance(value, LogLevelOption):
            raise TypeError(
                f"{self.__class__.__name__} expects a LogLevelOption value, but got {value!r}."
            )

        super().__init__(value)

    @classmethod
    def debug(cls) -> LogLevel:
        return LogLevel(LogLevelOption.Debug)

    @classmethod
    def error(cls) -> LogLevel:
        return LogLevel(LogLevelOption.Error)

    @classmethod
    def info(cls) -> LogLevel:
        return LogLevel(LogLevelOption.Info)

    def __str__(self) -> str:
        if self.value == LogLevelOption.Debug:
            return "DEBUG"
        elif self.value == LogLevelOption.Info:
            return "INFO"
        elif self.value == LogLevelOption.Error:
            return "ERROR"
        else:
            raise ValueError(f"{self.value} is not a LogLevelOption value.")


class LogMessage(ValueObject):
    def __init__(self, value: str, /):
        if not value:
            raise ValueError(
                f"{self.__class__.__name__} value is required, but got {value!r}."
            )

        value = str(value)
        if len(value) > 2000:
            warnings.warn(
                f"{self.__class__.__name__} must be <= 2000 characters long, but the message "
                f"is {len(value)}. It has been truncated to fit."
            )
            value = value[-2000:]

        super().__init__(value)


class SecondsSinceLastRefresh(ValueObject):
    def __init__(self, value: typing.Optional[int], /):
        if value is None:
            ...
        elif isinstance(value, int):
            if not value:
                raise ValueError(
                    f"If a {self.__class__.__name__} value is provided, then it must be positive, "
                    f"but got {value!r}."
                )
        else:
            raise TypeError(
                f"{self.__class__.__name__} expects an int, but got {value!r}"
            )

        super().__init__(value)



"""lime-etl barrel file"""
from lime_etl.domain import *  # noqa
from lime_etl.adapter import *  # noqa
from lime_etl.service import *  # noqa
from lime_uow.exceptions import *  # noqa
from lime_uow.unit_of_work import *  # noqa
from lime_uow.shared_resource_manager import *  # noqa
from lime_uow.resource import *  # noqa
from lime_uow.resources import *  # noqa



from __future__ import annotations

import datetime
import dataclasses

import typing

from lime_etl.domain import exceptions, job_result, value_objects


__all__ = (
    "BatchStatusDTO",
    "BatchStatus",
)


@dataclasses.dataclass(unsafe_hash=True)
class BatchStatusDTO:
    id: str
    name: str
    execution_error_message: typing.Optional[str]
    execution_error_occurred: typing.Optional[bool]
    execution_millis: typing.Optional[int]
    job_results: typing.List[job_result.JobResultDTO]
    running: bool
    ts: datetime.datetime

    def to_domain(self) -> BatchStatus:
        results = frozenset(job.to_domain() for job in self.job_results)
        if self.running:
            execution_millis = None
            execution_success_or_failure = None
        else:
            execution_millis = value_objects.ExecutionMillis(self.execution_millis or 0)
            if self.execution_error_occurred:
                execution_success_or_failure = value_objects.Result.failure(
                    self.execution_error_message or "No error message was provided."
                )
            else:
                execution_success_or_failure = value_objects.Result.success()

        return BatchStatus(
            id=value_objects.UniqueId(self.id),
            name=value_objects.BatchName(self.name),
            execution_millis=execution_millis,
            job_results=results,
            execution_success_or_failure=execution_success_or_failure,
            running=value_objects.Flag(self.running),
            ts=value_objects.Timestamp(self.ts),
        )


@dataclasses.dataclass(frozen=True)
class BatchStatus:
    id: value_objects.UniqueId
    name: value_objects.BatchName
    job_results: typing.FrozenSet[job_result.JobResult]
    execution_success_or_failure: typing.Optional[value_objects.Result]
    execution_millis: typing.Optional[value_objects.ExecutionMillis]
    running: value_objects.Flag
    ts: value_objects.Timestamp

    def __post_init__(self) -> None:
        if self.running.value is True:
            if self.execution_success_or_failure:
                raise exceptions.InvalidBatch(
                    f"If a batch is still running, execution_success_or_failure should be None, "
                    f"but got {self.execution_success_or_failure!r}."
                )
            if self.execution_millis:
                raise exceptions.InvalidBatch(
                    f"If a batch is running, execution_millis should be None, but got "
                    f"{self.execution_millis!r}."
                )
        else:
            if self.execution_success_or_failure is None:
                raise exceptions.InvalidBatch(
                    "If a bach has finished, then we should know the result, but "
                    "execution_success_or_failure is None."
                )
            if self.execution_millis is None:
                raise exceptions.InvalidBatch(
                    "If a batch has finished, then we should know how many milliseconds it took to "
                    "run, but execution_millis is None."
                )

    @property
    def job_names(self) -> typing.Set[value_objects.JobName]:
        return {job.job_name for job in self.job_results}

    @property
    def broken_jobs(self) -> typing.Set[value_objects.JobName]:
        return {job.job_name for job in self.job_results if job.tests_failed}

    def to_dto(self) -> BatchStatusDTO:
        results = [j.to_dto() for j in self.job_results]
        if self.execution_success_or_failure is None:
            error_occurred = None
            error_msg = None
        else:
            error_occurred = self.execution_success_or_failure.is_failure
            error_msg = self.execution_success_or_failure.failure_message_or_none

        if self.execution_millis is None:
            execution_millis = None
        else:
            execution_millis = self.execution_millis.value

        return BatchStatusDTO(
            id=self.id.value,
            name=self.name.value,
            execution_millis=execution_millis,
            running=self.running.value,
            job_results=results,
            execution_error_occurred=error_occurred,
            execution_error_message=error_msg,
            ts=self.ts.value,
        )



import datetime
import typing

from lime_uow import sqlalchemy_resources as lsa
from sqlalchemy import orm

from lime_etl import domain


__all__ = ("SqlAlchemyBatchLogRepository",)


class SqlAlchemyBatchLogRepository(
    domain.BatchLogRepository,
    lsa.SqlAlchemyRepository[domain.BatchLogEntryDTO],
):
    def __init__(
        self,
        session: orm.Session,
        ts_adapter: domain.TimestampAdapter,
    ):
        self._ts_adapter = ts_adapter
        super().__init__(session)

    def delete_old_entries(self, days_to_keep: domain.DaysToKeep) -> int:
        ts = self._ts_adapter.now().value
        cutoff = ts - datetime.timedelta(days=days_to_keep.value)
        return (
            self.session.query(domain.BatchLogEntryDTO)
            .filter(domain.BatchLogEntryDTO.ts < cutoff)
            .delete()
        )

    @property
    def entity_type(self) -> typing.Type[domain.BatchLogEntryDTO]:
        return domain.BatchLogEntryDTO

    def get_earliest_timestamp(self) -> typing.Optional[datetime.datetime]:
        earliest_entry = (
            self.session.query(domain.BatchLogEntryDTO)
            .order_by(domain.BatchLogEntryDTO.ts)
            .first()
        )
        if earliest_entry is None:
            return None
        else:
            return earliest_entry.ts

    @classmethod
    def interface(cls) -> typing.Type[domain.BatchLogRepository]:
        return domain.BatchLogRepository



import typing

from lime_etl.domain import job_dependency_errors, value_objects


__all__ = (
    "LimeETLException",
    "BatchNotFound",
    "DependencyErrors",
    "DuplicateJobNamesError",
    "InvalidBatch",
    "InvalidJobResult",
    "InvalidJobSpec",
    "InvalidResource",
    "MissingResourcesError",
)


class LimeETLException(Exception):
    """Base class for exceptions arising from the lime-etl package"""

    def __init__(self, message: str, /):
        self.message = message
        super().__init__(message)


class BatchNotFound(LimeETLException):
    def __init__(self, message: str, /):
        super().__init__(message)


class DependencyErrors(LimeETLException):
    def __init__(
        self,
        dependency_errors: typing.Set[job_dependency_errors.JobDependencyErrors],
        /,
    ):
        self.dependency_errors = dependency_errors
        # noinspection PyTypeChecker
        msg = "; ".join(str(e) for e in sorted(dependency_errors))
        super().__init__(msg)


class DuplicateJobNamesError(LimeETLException):
    def __init__(
        self, duplicate_job_counts: typing.Dict[value_objects.JobName, int], /
    ):
        self.duplicate_job_counts = duplicate_job_counts
        dupes_msg = ", ".join(
            f"[{job_name.value}] ({ct})"
            for job_name, ct in duplicate_job_counts.items()
        )
        err_msg = f"The following job names were included more than once: {dupes_msg}."
        super().__init__(err_msg)


class InvalidBatch(LimeETLException):
    def __init__(self, message: str, /):
        super().__init__(message)


class InvalidJobResult(LimeETLException):
    def __init__(self, message: str, /):
        super().__init__(message)


class InvalidJobSpec(LimeETLException):
    def __init__(self, message: str, /):
        super().__init__(message)


class InvalidResource(LimeETLException):
    def __init__(self, message: str, /):
        super().__init__(message)


class MissingResourcesError(LimeETLException):
    def __init__(
        self,
        missing_resources: typing.Mapping[
            value_objects.JobName, typing.Iterable[value_objects.ResourceName]
        ],
    ):
        self.missing_resources = missing_resources
        messages = "; ".join(
            f"[{job_name}] is missing {', '.join(sorted(f'[{m}]' for m in missing))}"
            for job_name, missing in missing_resources.items()
        )
        msg = f"The following jobs have missing resources: {messages}."
        super().__init__(msg)



import abc
import typing

import lime_uow as lu

from lime_etl.domain import batch_delta, batch_status, value_objects

__all__ = ("BatchRepository",)


class BatchRepository(lu.Repository[batch_status.BatchStatusDTO], abc.ABC):
    @abc.abstractmethod
    def delete_old_entries(self, days_to_keep: value_objects.DaysToKeep, /) -> int:
        raise NotImplementedError

    @abc.abstractmethod
    def get_latest(self, /, batch_name: value_objects.BatchName) -> typing.Optional[batch_status.BatchStatusDTO]:
        raise NotImplementedError

    @abc.abstractmethod
    def get_latest_batch_delta(self, /, batch_name: value_objects.BatchName) -> batch_delta.BatchDelta:
        raise NotImplementedError



from __future__ import annotations

import abc
import typing

import lime_uow as lu

from lime_etl.domain import admin_unit_of_work, job_logger, job_status, job_test_result, value_objects

__all__ = ("JobSpec",)

UoW = typing.TypeVar("UoW", bound=lu.UnitOfWork, contravariant=True)


class JobSpec(abc.ABC, typing.Generic[UoW]):
    @property
    def dependencies(self) -> typing.Tuple[value_objects.JobName, ...]:
        return tuple()

    @property
    @abc.abstractmethod
    def job_name(self) -> value_objects.JobName:
        raise NotImplementedError

    @property
    def min_seconds_between_refreshes(self) -> value_objects.MinSecondsBetweenRefreshes:
        return value_objects.MinSecondsBetweenRefreshes(0)

    @property
    def min_seconds_between_tests(self) -> value_objects.MinSecondsBetweenTests:
        return value_objects.MinSecondsBetweenTests(0)

    @property
    def max_retries(self) -> value_objects.MaxRetries:
        return value_objects.MaxRetries(0)

    def on_execution_error(self, error_message: str) -> typing.Optional[JobSpec[UoW]]:
        return None

    def on_test_failure(
        self, test_results: typing.FrozenSet[job_test_result.JobTestResult]
    ) -> typing.Optional[JobSpec[UoW]]:
        return None

    @abc.abstractmethod
    def run(
        self,
        admin_uow: admin_unit_of_work.AdminUnitOfWork,
        batch_uow: UoW,
        logger: job_logger.JobLogger,
    ) -> job_status.JobStatus:
        raise NotImplementedError

    def test(
        self,
        admin_uow: admin_unit_of_work.AdminUnitOfWork,
        batch_uow: UoW,
        logger: job_logger.JobLogger,
    ) -> typing.List[job_test_result.SimpleJobTestResult]:
        return []

    @property
    def timeout_seconds(self) -> value_objects.TimeoutSeconds:
        return value_objects.TimeoutSeconds(None)

    def __repr__(self) -> str:
        return f"<JobSpec: {self.__class__.__name__}>: {self.job_name.value}"

    def __hash__(self) -> int:
        return hash(self.job_name.value)

    def __eq__(self, other: object) -> bool:
        if other.__class__ is self.__class__:
            return (
                self.job_name.value == typing.cast(JobSpec[UoW], other).job_name.value
            )
        else:
            return NotImplemented



import abc

import lime_uow as lu

from lime_etl.domain import value_objects

__all__ = ("TimestampAdapter",)


class TimestampAdapter(lu.Resource[None], abc.ABC):
    @abc.abstractmethod
    def now(self) -> value_objects.Timestamp:
        raise NotImplementedError

    def get_elapsed_time(self, start_ts: value_objects.Timestamp) -> value_objects.ExecutionMillis:
        end_ts = self.now()
        millis = int((end_ts.value - start_ts.value).total_seconds() * 1000)
        return value_objects.ExecutionMillis(millis)



from sqlalchemy import (
    Boolean,
    Column,
    DateTime,
    Enum,
    ForeignKey,
    Integer,
    MetaData,
    String,
    Table,
)
from sqlalchemy.orm import mapper, relationship
from sqlalchemy.orm.base import _is_mapped_class  # type: ignore

from lime_etl.domain import (
    batch_status,
    job_result,
    job_log_entry,
    job_test_result,
    batch_log_entry,
    value_objects,
)


__all__ = (
    "admin_metadata",
    "set_schema",
    "start_mappers",
)

admin_metadata = MetaData()


batches = Table(
    "batches",
    admin_metadata,
    Column("id", String(32), primary_key=True),
    Column("name", String(200), nullable=False),
    Column("execution_millis", Integer, nullable=True),
    Column("execution_error_occurred", Boolean, nullable=True),
    Column("execution_error_message", String(2000), nullable=True),
    Column("running", Boolean, nullable=False),
    Column("ts", DateTime, nullable=False),
)

batch_log = Table(
    "batch_log",
    admin_metadata,
    Column("id", String(32), primary_key=True),
    Column("batch_id", String(32), nullable=False),
    Column("log_level", Enum(value_objects.LogLevelOption)),
    Column("message", String(2000), nullable=False),
    Column("ts", DateTime, nullable=False),
)

job_log = Table(
    "job_log",
    admin_metadata,
    Column("id", String(32), primary_key=True),
    Column("batch_id", String(32), nullable=False),
    Column("job_id", String(32), nullable=False),
    Column("log_level", Enum(value_objects.LogLevelOption)),
    Column("message", String(2000), nullable=False),
    Column("ts", DateTime, nullable=False),
)

jobs = Table(
    "jobs",
    admin_metadata,
    Column("id", String(32), primary_key=True),
    Column("batch_id", ForeignKey("batches.id")),
    Column("job_name", String(200), nullable=False),
    Column("execution_millis", Integer, nullable=True),
    Column("execution_error_occurred", Boolean, nullable=True),
    Column("execution_error_message", String(2000), nullable=True),
    Column("running", Boolean, nullable=False),
    Column("skipped", Boolean, nullable=False),
    Column("ts", DateTime, nullable=False),
)

job_test_results = Table(
    "job_test_results",
    admin_metadata,
    Column("id", String(32), primary_key=True),
    Column("job_id", ForeignKey("jobs.id")),
    Column("test_name", String(200), nullable=False),
    Column("test_passed", Boolean, nullable=True),
    Column("test_failure_message", String(2000), nullable=True),
    Column("execution_millis", Integer, nullable=False),
    Column("execution_error_occurred", Boolean, nullable=False),
    Column("execution_error_message", String(2000), nullable=True),
    Column("ts", DateTime, nullable=False),
)


def set_schema(schema: value_objects.SchemaName) -> None:
    for table_name, table in admin_metadata.tables.items():
        table.schema = schema.value


def start_mappers() -> None:
    if not _is_mapped_class(batch_log_entry.BatchLogEntryDTO):
        mapper(batch_log_entry.BatchLogEntryDTO, batch_log)
        mapper(job_log_entry.JobLogEntryDTO, job_log)
        job_test_result_mapper = mapper(
            job_test_result.JobTestResultDTO, job_test_results
        )
        job_mapper = mapper(
            job_result.JobResultDTO,
            jobs,
            properties={
                "test_results": relationship(
                    job_test_result_mapper,
                    cascade="all,delete,delete-orphan",
                    collection_class=list,
                ),
            },
        )
        mapper(
            batch_status.BatchStatusDTO,
            batches,
            properties={
                "job_results": relationship(
                    job_mapper,
                    cascade="all,delete,delete-orphan",
                    collection_class=list,
                ),
            },
        )



from sqlalchemy import orm

from lime_etl import domain
from lime_etl.adapter import sqlalchemy_job_logger

__all__ = (
    "SqlAlchemyBatchLogger",
    "ConsoleBatchLogger",
)


class SqlAlchemyBatchLogger(domain.BatchLogger):
    def __init__(
        self,
        *,
        batch_id: domain.UniqueId,
        session: orm.Session,
        ts_adapter: domain.TimestampAdapter,
    ):
        self._batch_id = batch_id
        self._session = session
        self._ts_adapter = ts_adapter
        super().__init__()

    def create_job_logger(self, /, job_id: domain.UniqueId) -> domain.JobLogger:
        return sqlalchemy_job_logger.SqlAlchemyJobLogger(
            batch_id=self._batch_id,
            job_id=job_id,
            session=self._session,
            ts_adapter=self._ts_adapter,
        )

    def _log(self, level: domain.LogLevel, message: str) -> None:
        log_entry = domain.BatchLogEntry(
            id=domain.UniqueId.generate(),
            batch_id=self._batch_id,
            log_level=level,
            message=domain.LogMessage(message),
            ts=self._ts_adapter.now(),
        )
        self._session.add(log_entry.to_dto())
        self._session.commit()
        return None

    def log_error(self, message: str, /) -> None:
        return self._log(
            level=domain.LogLevel.error(),
            message=message,
        )

    def log_info(self, message: str, /) -> None:
        return self._log(
            level=domain.LogLevel.info(),
            message=message,
        )


class ConsoleBatchLogger(domain.BatchLogger):
    def __init__(self, batch_id: domain.UniqueId):
        self.batch_id = batch_id
        super().__init__()

    def create_job_logger(self, /, job_id: domain.UniqueId) -> domain.JobLogger:
        return sqlalchemy_job_logger.ConsoleJobLogger()

    def log_error(self, message: str, /) -> None:
        print(f"ERROR: {message}")
        return None

    def log_info(self, message: str, /) -> None:
        print(f"INFO: {message}")
        return None



from __future__ import annotations

import abc
import datetime
import typing

import lime_uow as lu

from lime_etl.domain import batch_log_entry, value_objects

__all__ = ("BatchLogRepository",)


class BatchLogRepository(
    lu.Repository[batch_log_entry.BatchLogEntryDTO],
    abc.ABC,
):
    @abc.abstractmethod
    def delete_old_entries(self, days_to_keep: value_objects.DaysToKeep) -> int:
        raise NotImplementedError

    @abc.abstractmethod
    def get_earliest_timestamp(self) -> typing.Optional[datetime.datetime]:
        raise NotImplementedError



import typing

import sqlalchemy as sa
from sqlalchemy import orm

from lime_etl import domain, adapter
from lime_etl.service import admin, batch_runner

__all__ = ("AdminBatch",)


class AdminBatch(domain.BatchSpec[domain.admin_unit_of_work.AdminUnitOfWork]):
    def __init__(
        self,
        *,
        admin_engine_uri: domain.DbUri,
        admin_schema: domain.SchemaName = domain.SchemaName("etl"),
        days_logs_to_keep: domain.DaysToKeep = domain.DaysToKeep(3),
        min_seconds_between_runs: domain.MinSecondsBetweenRefreshes = domain.MinSecondsBetweenRefreshes(12 * 60 * 60),
        ts_adapter: domain.TimestampAdapter = adapter.LocalTimestampAdapter(),
    ):
        self._admin_engine_uri = admin_engine_uri
        self._admin_schema = admin_schema
        self._days_logs_to_keep = days_logs_to_keep
        self._min_seconds_between_runs = min_seconds_between_runs
        self._ts_adapter = ts_adapter

    @property
    def batch_name(self) -> domain.BatchName:
        return domain.BatchName("admin")

    def create_jobs(
        self, uow: domain.admin_unit_of_work.AdminUnitOfWork
    ) -> typing.List[domain.JobSpec[domain.admin_unit_of_work.AdminUnitOfWork]]:
        return [
            admin.delete_old_logs.DeleteOldLogs(
                days_logs_to_keep=self._days_logs_to_keep,
                min_seconds_between_runs=self._min_seconds_between_runs,
            ),
        ]

    def create_uow(self) -> domain.admin_unit_of_work.AdminUnitOfWork:
        admin_engine = sa.create_engine(self._admin_engine_uri.value)
        adapter.admin_metadata.create_all(bind=admin_engine)
        adapter.admin_orm.set_schema(schema=self._admin_schema)
        adapter.admin_orm.start_mappers()
        admin_session_factory = orm.sessionmaker(bind=admin_engine)
        return adapter.SqlAlchemyAdminUnitOfWork(
            session_factory=admin_session_factory, ts_adapter=self._ts_adapter
        )

    def run(self) -> domain.BatchStatus:
        return batch_runner.run_batch(
            batch=self,
            admin_engine_uri=self._admin_engine_uri,
            admin_schema=self._admin_schema,
            ts_adapter=self._ts_adapter,
        )



from __future__ import annotations

import dataclasses
import datetime
import typing

from lime_etl.domain import job_status, job_test_result, value_objects

__all__ = (
    "JobResultDTO",
    "JobResult",
)


@dataclasses.dataclass(unsafe_hash=True)
class JobResultDTO:
    id: str
    batch_id: str
    job_name: str
    test_results: typing.List[job_test_result.JobTestResultDTO]
    execution_millis: typing.Optional[int]
    execution_error_occurred: typing.Optional[bool]
    execution_error_message: typing.Optional[str]
    running: bool
    skipped: bool
    ts: datetime.datetime

    def to_domain(self) -> JobResult:
        test_results = frozenset(dto.to_domain() for dto in self.test_results)
        if self.running:
            execution_millis = None
            status: job_status.JobStatus = job_status.JobStatus.in_progress()
        else:
            execution_millis = value_objects.ExecutionMillis(self.execution_millis or 0)
            if self.execution_error_occurred:
                status = job_status.JobStatus.failed(
                    self.execution_error_message or "No error message was provided."
                )
            else:
                status = job_status.JobStatus.success()

        return JobResult(
            id=value_objects.UniqueId(self.id),
            batch_id=value_objects.UniqueId(self.batch_id),
            job_name=value_objects.JobName(self.job_name),
            test_results=test_results,
            execution_millis=execution_millis,
            status=status,
            ts=value_objects.Timestamp(self.ts),
        )


@dataclasses.dataclass(frozen=True)
class JobResult:
    batch_id: value_objects.UniqueId
    execution_millis: typing.Optional[value_objects.ExecutionMillis]
    status: job_status.JobStatus
    id: value_objects.UniqueId
    job_name: value_objects.JobName
    ts: value_objects.Timestamp
    test_results: typing.FrozenSet[job_test_result.JobTestResult]

    @property
    def tests_failed(self) -> bool:
        return any(result.test_failed for result in self.test_results)

    def to_dto(self) -> JobResultDTO:
        test_results = [r.to_dto() for r in self.test_results]

        if isinstance(self.status, job_status.JobFailed):
            error_occurred: typing.Optional[bool] = True
            error_message: typing.Optional[str] = self.status.error_message.value
            running = False
            skipped = False
        elif isinstance(self.status, job_status.JobInProgress):
            error_occurred = None
            error_message = None
            running = True
            skipped = False
        elif isinstance(self.status, job_status.JobSkipped):
            error_occurred = False
            error_message = None
            running = False
            skipped = True
        elif isinstance(self.status, job_status.JobRanSuccessfully):
            error_occurred = False
            error_message = None
            running = False
            skipped = False
        else:
            raise ValueError(f"Expected an instance of job_status.JobStatus, but got {self.status!r}.")

        if self.execution_millis is None:
            execution_millis = None
        else:
            execution_millis = self.execution_millis.value

        return JobResultDTO(
            id=self.id.value,
            batch_id=self.batch_id.value,
            job_name=self.job_name.value,
            test_results=test_results,
            execution_millis=execution_millis,
            execution_error_occurred=error_occurred,
            execution_error_message=error_message,
            running=running,
            skipped=skipped,
            ts=self.ts.value,
        )

    @staticmethod
    def running(
        *,
        job_status_id: value_objects.UniqueId,
        batch_id: value_objects.UniqueId,
        job_name: value_objects.JobName,
        ts: value_objects.Timestamp,
    ) -> JobResult:
        return JobResult(
            id=job_status_id,
            batch_id=batch_id,
            job_name=job_name,
            ts=ts,
            test_results=frozenset(),
            execution_millis=None,
            status=job_status.JobStatus.in_progress(),
        )



import dataclasses

import typing

from lime_etl.domain import batch_status, value_objects


@dataclasses.dataclass(frozen=True)
class BatchDelta:
    current_results: batch_status.BatchStatus
    previous_results: typing.Optional[batch_status.BatchStatus]

    @property
    def common_jobs(self) -> typing.Set[value_objects.JobName]:
        if self.previous_results is None:
            return set()
        else:
            return self.previous_results.job_names & self.current_results.job_names

    @property
    def newly_broken_jobs(self) -> typing.Set[value_objects.JobName]:
        if self.previous_results is None:
            return self.current_results.broken_jobs
        else:
            return self.current_results.broken_jobs - self.previous_results.broken_jobs

    @property
    def newly_fixed_jobs(self) -> typing.Set[value_objects.JobName]:
        if self.previous_results is None:
            return self.current_results.broken_jobs
        else:
            return self.previous_results.broken_jobs - self.current_results.broken_jobs

    def __str__(self) -> str:
        return (
            "The following jobs are broken:\n    " + "\n    ".join(f"{j}\n    " for j in self.current_results.broken_jobs) +
            "The following jobs are newly broken:\n    " + "\n    ".join(f"{j}\n    " for j in self.newly_broken_jobs) +
            "The following jobs have been fixed:\n    " + "\n    ".join(f"{j}\n    " for j in self.newly_fixed_jobs)
        )



from __future__ import annotations

import typing

import lime_uow as lu
from sqlalchemy import orm

from lime_etl import domain
from lime_etl.adapter import (
    local_timestamp_adapter,
    sqlalchemy_admin_session,
    sqlalchemy_batch_repository,
    sqlalchemy_batch_log_repository,
    sqlalchemy_job_log_repository,
    sqlalchemy_job_repository,
)

__all__ = ("SqlAlchemyAdminUnitOfWork",)


class SqlAlchemyAdminUnitOfWork(domain.AdminUnitOfWork):
    def __init__(
        self,
        session_factory: orm.sessionmaker,
        ts_adapter: domain.TimestampAdapter = local_timestamp_adapter.LocalTimestampAdapter(),
    ):
        super().__init__()
        self._session_factory = session_factory
        self._ts_adapter = ts_adapter

    @property
    def batch_repo(self) -> domain.BatchRepository:
        return self.get(domain.BatchRepository)  # type: ignore  # see mypy issue 5374

    @property
    def batch_log_repo(self) -> domain.BatchLogRepository:
        return self.get(domain.BatchLogRepository)  # type: ignore  # see mypy issue 5374

    def create_shared_resources(self) -> typing.List[lu.Resource[typing.Any]]:
        return [sqlalchemy_admin_session.SqlAlchemyAdminSession(self._session_factory)]

    @property
    def job_repo(self) -> domain.JobRepository:
        return self.get(domain.JobRepository)  # type: ignore  # see mypy issue 5374

    @property
    def job_log_repo(self) -> domain.JobLogRepository:
        return self.get(domain.JobLogRepository)  # type: ignore  # see mypy issue 5374

    @property
    def ts_adapter(self) -> domain.TimestampAdapter:
        return self._ts_adapter

    def create_resources(
        self, shared_resources: lu.SharedResourceManager
    ) -> typing.Set[lu.Resource[typing.Any]]:
        session = shared_resources.get(sqlalchemy_admin_session.SqlAlchemyAdminSession)
        return {
            sqlalchemy_batch_repository.SqlAlchemyBatchRepository(
                session=session, ts_adapter=self._ts_adapter
            ),
            sqlalchemy_batch_log_repository.SqlAlchemyBatchLogRepository(
                session=session, ts_adapter=self._ts_adapter
            ),
            sqlalchemy_job_repository.SqlAlchemyJobRepository(
                session=session, ts_adapter=self._ts_adapter
            ),
            sqlalchemy_job_log_repository.SqlAlchemyJobLogRepository(
                session=session, ts_adapter=self._ts_adapter
            ),
        }



import abc

import lime_uow as lu

from lime_etl.domain import (
    batch_repository,
    batch_log_repository,
    job_repository,
    job_log_repository,
    timestamp_adapter,
)

__all__ = ("AdminUnitOfWork",)


class AdminUnitOfWork(lu.UnitOfWork, abc.ABC):
    @property
    @abc.abstractmethod
    def batch_repo(self) -> batch_repository.BatchRepository:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def batch_log_repo(self) -> batch_log_repository.BatchLogRepository:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def job_repo(self) -> job_repository.JobRepository:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def job_log_repo(self) -> job_log_repository.JobLogRepository:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def ts_adapter(self) -> timestamp_adapter.TimestampAdapter:
        raise NotImplementedError



import datetime
import itertools
import multiprocessing
import traceback
import typing

import lime_uow as lu
import sqlalchemy as sa
from sqlalchemy import orm

from lime_etl import domain, adapter
from lime_etl.service import admin

__all__ = (
    "run_admin",
    "run_batch",
    "run_batches_in_parallel",
)

UoW = typing.TypeVar("UoW", bound=lu.UnitOfWork, contravariant=True)


def run_admin(
    *,
    admin_engine_uri: domain.DbUri,
    admin_schema: domain.SchemaName = domain.SchemaName("etl"),
    days_logs_to_keep: domain.DaysToKeep = domain.DaysToKeep(3),
    ts_adapter: domain.TimestampAdapter = adapter.LocalTimestampAdapter(),
) -> domain.BatchStatus:
    batch = admin.AdminBatch(
        admin_engine_uri=admin_engine_uri,
        admin_schema=admin_schema,
        days_logs_to_keep=days_logs_to_keep,
        ts_adapter=ts_adapter,
    )
    return batch.run()


def run_batches_in_parallel(
    batches: typing.Iterable[domain.BatchSpec[UoW]],
    admin_engine_uri: domain.DbUri,
    admin_schema: domain.SchemaName = domain.SchemaName("etl"),
    max_processes: domain.MaxProcesses = domain.MaxProcesses(None),
    timeout: domain.TimeoutSeconds = domain.TimeoutSeconds(None),
    ts_adapter: domain.TimestampAdapter = adapter.LocalTimestampAdapter(),
) -> typing.List[domain.BatchStatus]:
    params = [(batch, admin_engine_uri, admin_schema, ts_adapter) for batch in batches]
    with multiprocessing.Pool(max_processes.value, maxtasksperchild=1) as pool:
        future = pool.starmap_async(run_batch, params)
        return future.get(timeout.value)


def run_batch(
    batch: domain.BatchSpec[UoW],
    admin_engine_uri: domain.DbUri,
    admin_schema: domain.SchemaName = domain.SchemaName("etl"),
    ts_adapter: domain.TimestampAdapter = adapter.LocalTimestampAdapter(),
) -> domain.BatchStatus:
    start_time = ts_adapter.now()

    admin_engine = sa.create_engine(admin_engine_uri.value)
    adapter.admin_metadata.create_all(bind=admin_engine)
    adapter.admin_orm.set_schema(schema=admin_schema)
    adapter.admin_orm.start_mappers()
    admin_session_factory = orm.sessionmaker(bind=admin_engine)

    logger = adapter.SqlAlchemyBatchLogger(
        batch_id=batch.batch_id,
        session=admin_session_factory(),
        ts_adapter=ts_adapter,
    )
    admin_uow: domain.AdminUnitOfWork = adapter.SqlAlchemyAdminUnitOfWork(
        session_factory=admin_session_factory, ts_adapter=ts_adapter
    )

    try:
        with admin_uow as uow:
            new_batch = domain.BatchStatus(
                id=batch.batch_id,
                name=batch.batch_name,
                job_results=frozenset(),
                execution_millis=None,
                execution_success_or_failure=None,
                running=domain.Flag(True),
                ts=start_time,
            )
            uow.batch_repo.add(new_batch.to_dto())
            uow.save()

        logger.log_info(f"Staring batch [{batch.batch_name.value}]...")
        batch_uow = batch.create_uow()
        try:
            result = run_batch_or_fail(
                admin_uow=admin_uow,
                batch=batch,
                batch_uow=batch_uow,
                logger=logger,
                start_time=start_time,
                ts_adapter=ts_adapter,
            )
        finally:
            batch_uow.close()

        with admin_uow as uow:
            uow.batch_repo.update(result.to_dto())
            uow.save()

        logger.log_info(f"Batch [{batch.batch_name}] finished.")
        return result
    except Exception as e:
        logger.log_error(traceback.format_exc(10))
        end_time = ts_adapter.now()
        with admin_uow as uow:
            result = domain.BatchStatus(
                id=batch.batch_id,
                name=batch.batch_name,
                job_results=frozenset(),
                execution_success_or_failure=domain.Result.failure(f"{e}\n{traceback.format_exc(10)}"),
                execution_millis=domain.ExecutionMillis.calculate(
                    start_time=start_time, end_time=end_time
                ),
                running=domain.Flag(False),
                ts=start_time,
            )
            uow.batch_repo.update(result.to_dto())
            uow.save()
        raise
    finally:
        admin_uow.close()


def run_batch_or_fail(
    *,
    admin_uow: domain.admin_unit_of_work.AdminUnitOfWork,
    batch: domain.BatchSpec[UoW],
    batch_uow: UoW,
    logger: domain.batch_logger.BatchLogger,
    start_time: domain.Timestamp,
    ts_adapter: domain.TimestampAdapter,
) -> domain.BatchStatus:
    jobs = batch.create_jobs(batch_uow)
    check_dependencies(jobs)
    check_for_duplicate_job_names(jobs)

    job_results: typing.List[domain.JobResult] = []
    for ix, job in enumerate(jobs):
        job_id = domain.UniqueId.generate()

        if job.dependencies and all(
            isinstance(r.status, (domain.JobSkipped, domain.JobFailed))
            for r in job_results
            if r.job_name in job.dependencies
        ):
            logger.log_info(
                f"All the dependencies for [{job.job_name.value}] were skipped or failed so the job has been skipped."
            )
            result = domain.JobResult(
                id=job_id,
                batch_id=batch.batch_id,
                job_name=job.job_name,
                test_results=frozenset(),
                execution_millis=domain.ExecutionMillis(0),
                status=domain.JobStatus.skipped("Dependencies were skipped or failed."),
                ts=start_time,
            )
        else:
            current_ts = ts_adapter.now()
            with admin_uow as uow:
                last_ts = uow.job_repo.get_last_successful_ts(job.job_name)

            if last_ts:
                seconds_since_last_refresh = (
                    current_ts.value - last_ts.value
                ).total_seconds()
                time_to_run_again = (
                    seconds_since_last_refresh > job.min_seconds_between_refreshes.value
                )
            else:
                seconds_since_last_refresh = 0
                time_to_run_again = True

            if time_to_run_again:
                job_logger = logger.create_job_logger(job_id)
                result = domain.JobResult.running(
                    job_status_id=job_id,
                    batch_id=batch.batch_id,
                    job_name=job.job_name,
                    ts=start_time,
                )
                with admin_uow as uow:
                    uow.job_repo.add(result.to_dto())
                    uow.save()

                # noinspection PyBroadException
                try:
                    result = run_job(
                        admin_uow=admin_uow,
                        batch=batch,
                        batch_uow=batch_uow,
                        job=job,
                        logger=job_logger,
                        job_id=job_id,
                        ts_adapter=ts_adapter,
                    )
                except Exception as e:
                    millis = ts_adapter.get_elapsed_time(start_time)
                    err_msg = f"{e}\n{traceback.format_exc(10)}"
                    result = domain.JobResult(
                        id=job_id,
                        batch_id=batch.batch_id,
                        job_name=job.job_name,
                        test_results=frozenset(),
                        execution_millis=millis,
                        status=domain.JobStatus.failed(err_msg),
                        ts=result.ts,
                    )
            else:
                logger.log_info(
                    f"[{job.job_name.value}] was run successfully {seconds_since_last_refresh:.0f} seconds "
                    f"ago and it is set to refresh every {job.min_seconds_between_refreshes.value} seconds, "
                    f"so there is no need to refresh again."
                )
                result = domain.JobResult(
                    id=job_id,
                    batch_id=batch.batch_id,
                    job_name=job.job_name,
                    test_results=frozenset(),
                    execution_millis=domain.ExecutionMillis(0),
                    status=domain.JobStatus.skipped(
                        f"The job ran {seconds_since_last_refresh:.0f} seconds ago, so it is not time yet."
                    ),
                    ts=start_time,
                )

        job_results.append(result)
        with admin_uow as uow:
            uow.job_repo.update(result.to_dto())
            uow.save()

    end_time = ts_adapter.now()

    execution_millis = int((end_time.value - start_time.value).total_seconds() * 1000)
    return domain.BatchStatus(
        id=batch.batch_id,
        name=batch.batch_name,
        execution_millis=domain.ExecutionMillis(execution_millis),
        job_results=frozenset(job_results),
        execution_success_or_failure=domain.Result.success(),
        running=domain.Flag(False),
        ts=end_time,
    )


def run_job(
    *,
    admin_uow: domain.admin_unit_of_work.AdminUnitOfWork,
    batch: domain.BatchSpec[UoW],
    batch_uow: UoW,
    job: domain.JobSpec[UoW],
    job_id: domain.UniqueId,
    logger: domain.JobLogger,
    ts_adapter: domain.TimestampAdapter,
) -> domain.JobResult:
    result: domain.JobResult = run_job_pre_handlers(
        admin_uow=admin_uow,
        batch=batch,
        batch_uow=batch_uow,
        job=job,
        job_id=job_id,
        logger=logger,
        ts_adapter=ts_adapter,
    )
    assert result.status is not None
    if isinstance(result.status, domain.JobFailed):
        new_job = job.on_execution_error(result.status.error_message.value)
        if new_job:
            return run_job(
                admin_uow=admin_uow,
                batch=batch,
                batch_uow=batch_uow,
                job=new_job,
                job_id=job_id,
                logger=logger,
                ts_adapter=ts_adapter,
            )
        else:
            return result
    elif any(test.test_failed for test in result.test_results):
        new_job = job.on_test_failure(result.test_results)
        if new_job:
            return run_job(
                admin_uow=admin_uow,
                batch=batch,
                batch_uow=batch_uow,
                job=new_job,
                job_id=job_id,
                logger=logger,
                ts_adapter=ts_adapter,
            )
        else:
            return result
    else:
        return result


def run_job_pre_handlers(
    *,
    admin_uow: domain.admin_unit_of_work.AdminUnitOfWork,
    batch: domain.BatchSpec[UoW],
    batch_uow: UoW,
    job: domain.JobSpec[UoW],
    job_id: domain.UniqueId,
    logger: domain.JobLogger,
    ts_adapter: domain.TimestampAdapter,
) -> domain.JobResult:
    logger.log_info(f"Starting [{job.job_name.value}]...")
    start_time = ts_adapter.now()
    with admin_uow as uow:
        current_batch = uow.batch_repo.get(batch.batch_id.value).to_domain()

    if current_batch is None:
        msg = f"The batch [{batch.batch_id.value}] was not found"
        raise domain.exceptions.BatchNotFound(msg)

    dep_exceptions = {
        jr.job_name.value
        for jr in current_batch.job_results
        if jr.job_name in job.dependencies
        and jr.status is not None
        and isinstance(jr.status, domain.JobFailed)
    }
    dep_test_failures = {
        jr.job_name.value
        for jr in current_batch.job_results
        if jr.job_name in job.dependencies and jr.tests_failed
    }
    if dep_exceptions and dep_test_failures:
        errs = ", ".join(sorted(dep_exceptions))
        test_failures = ", ".join(sorted(dep_test_failures))
        raise Exception(
            f"The following dependencies failed to execute: {errs} "
            f"and the following jobs had test failures: {test_failures}"
        )
    elif dep_exceptions:
        errs = ", ".join(sorted(dep_exceptions))
        raise Exception(f"The following dependencies failed to execute: {errs}")
    else:
        result = run_jobs_with_tests(
            admin_uow=admin_uow,
            batch=batch,
            batch_uow=batch_uow,
            job=job,
            job_id=job_id,
            logger=logger,
            start_time=start_time,
            ts_adapter=ts_adapter,
        )
        logger.log_info(f"Finished running [{job.job_name.value}].")
        return result


def run_jobs_with_tests(
    *,
    admin_uow: domain.AdminUnitOfWork,
    batch: domain.BatchSpec[UoW],
    batch_uow: UoW,
    job: domain.JobSpec[UoW],
    job_id: domain.UniqueId,
    logger: domain.JobLogger,
    start_time: domain.Timestamp,
    ts_adapter: domain.TimestampAdapter,
) -> domain.JobResult:
    result, execution_millis = run_job_with_retry(
        admin_uow=admin_uow,
        batch=batch,
        batch_uow=batch_uow,
        job=job,
        logger=logger,
        max_retries=job.max_retries.value,
        retries_so_far=0,
        start_time=start_time,
        ts_adapter=ts_adapter,
    )
    if isinstance(result, domain.JobRanSuccessfully):
        logger.log_info(f"[{job.job_name.value}] finished successfully.")

        if batch.skip_tests.value:
            full_test_results: typing.FrozenSet[domain.JobTestResult] = frozenset()
        else:
            with admin_uow:
                last_test_results = admin_uow.job_repo.latest_test_results(job.job_name)

            if last_test_results:
                last_test_time: datetime.datetime = max(
                    t.ts.value for t in last_test_results
                )
                now: datetime.datetime = ts_adapter.now().value
                seconds_since_last_test_run: int = int(
                    (now - last_test_time).total_seconds()
                )
                if seconds_since_last_test_run >= job.min_seconds_between_tests.value:
                    skip_tests = False
                    logger.log_info(
                        f"The tests for [{job.job_name.value}] were last run {seconds_since_last_test_run} seconds "
                        f"ago, and they are set to run every {job.min_seconds_between_tests.value}, so they will "
                        f"be run again now."
                    )
                else:
                    skip_tests = True
                    logger.log_info(
                        f"The tests for [{job.job_name.value}] were run {seconds_since_last_test_run} seconds ago, "
                        f"and they are set to run every {job.min_seconds_between_tests.value} so they are not "
                        f"ready to be run again."
                    )
            else:
                skip_tests = False
                logger.log_info(
                    f"The tests for [{job.job_name.value}] have not been run before, so they will be run now."
                )

            if skip_tests:
                full_test_results = frozenset()
            else:
                test_start_time = datetime.datetime.now()
                test_results = job.test(admin_uow, batch_uow, logger)
                test_execution_millis = int(
                    (datetime.datetime.now() - test_start_time).total_seconds() * 1000
                )

                if test_results:
                    tests_passed = sum(
                        1 for test_result in test_results if test_result.test_passed
                    )
                    tests_failed = sum(
                        1 for test_result in test_results if test_result.test_failed
                    )
                    logger.log_info(
                        f"{job.job_name.value} test results: {tests_passed=}, {tests_failed=}"
                    )
                    full_test_results = frozenset(
                        domain.JobTestResult(
                            id=domain.UniqueId.generate(),
                            job_id=job_id,
                            test_name=test_result.test_name,
                            test_success_or_failure=test_result.test_success_or_failure,
                            execution_millis=domain.ExecutionMillis(test_execution_millis),
                            execution_success_or_failure=domain.Result.success(),
                            ts=start_time,
                        )
                        for test_result in test_results
                    )
                else:
                    logger.log_info("The job test method returned no results.")
                    full_test_results = frozenset()
    elif isinstance(result, domain.JobFailed):
        logger.log_info(
            f"An exception occurred while running [{job.job_name.value}]: "
            f"{result.error_message}."
        )
        full_test_results = frozenset()
    elif isinstance(result, domain.JobSkipped):
        logger.log_info(f"[{job.job_name.value}] was skipped.")
        full_test_results = frozenset()
    else:
        raise TypeError(f"Unrecognized job result: {result!r}")

    return domain.JobResult(
        id=job_id,
        batch_id=batch.batch_id,
        job_name=job.job_name,
        test_results=full_test_results,
        execution_millis=execution_millis,
        status=result,
        ts=start_time,
    )


def run_job_with_retry(
    *,
    admin_uow: domain.AdminUnitOfWork,
    batch: domain.BatchSpec[UoW],
    batch_uow: UoW,
    job: domain.JobSpec[UoW],
    logger: domain.JobLogger,
    max_retries: int,
    retries_so_far: int,
    start_time: domain.Timestamp,
    ts_adapter: domain.TimestampAdapter,
) -> typing.Tuple[domain.JobStatus, domain.ExecutionMillis]:
    # noinspection PyBroadException
    try:
        result = job.run(admin_uow, batch_uow, logger) or domain.JobStatus.success()
        end_time = ts_adapter.now()
        execution_millis = domain.ExecutionMillis.calculate(
            start_time=start_time, end_time=end_time
        )
        return result, execution_millis
    except:
        if max_retries > retries_so_far:
            logger.log_info(f"Running retry {retries_so_far} of {max_retries}...")
            return run_job_with_retry(
                admin_uow=admin_uow,
                batch=batch,
                batch_uow=batch_uow,
                job=job,
                logger=logger,
                max_retries=max_retries,
                retries_so_far=retries_so_far + 1,
                start_time=start_time,
                ts_adapter=ts_adapter,
            )
        else:
            logger.log_info(
                f"[{job.job_name.value}] failed after {max_retries} retries."
            )
            raise


def check_for_duplicate_job_names(
    jobs: typing.Collection[domain.JobSpec[UoW]], /
) -> None:
    job_names = [job.job_name for job in jobs]
    duplicates = {
        job_name: ct for job_name in job_names if (ct := job_names.count(job_name)) > 1
    }
    if duplicates:
        raise domain.exceptions.DuplicateJobNamesError(duplicates)


def check_dependencies(jobs: typing.List[domain.JobSpec[UoW]], /) -> None:
    job_names = {job.job_name for job in jobs}
    unresolved_dependencies_by_table = {
        job.job_name: set(dep for dep in job.dependencies if dep not in job_names)
        for job in jobs
        if any(dep not in job_names for dep in job.dependencies)
    }
    unresolved_dependencies = {
        dep for dep_grp in unresolved_dependencies_by_table.values() for dep in dep_grp
    }

    job_names_seen_so_far: typing.List[domain.JobName] = []
    jobs_out_of_order_by_table: typing.Dict[
        domain.JobName, typing.Set[domain.JobName]
    ] = dict()
    for job in jobs:
        job_names_seen_so_far.append(job.job_name)
        job_deps_out_of_order = []
        for dep in job.dependencies:
            if dep not in job_names_seen_so_far and dep not in unresolved_dependencies:
                job_deps_out_of_order.append(dep)
        if job_deps_out_of_order:
            jobs_out_of_order_by_table[job.job_name] = set(job_deps_out_of_order)

    dependency_errors = {
        domain.JobDependencyErrors(
            job_name=job_name,
            missing_dependencies=frozenset(
                unresolved_dependencies_by_table.get(job_name, set())
            ),
            jobs_out_of_order=frozenset(
                jobs_out_of_order_by_table.get(job_name, set())
            ),
        )
        for job_name in set(
            itertools.chain(
                unresolved_dependencies_by_table.keys(),
                jobs_out_of_order_by_table.keys(),
            )
        )
    }
    if dependency_errors:
        raise domain.exceptions.DependencyErrors(dependency_errors)



from lime_etl.adapter.admin_orm import *
from lime_etl.adapter.sqlalchemy_admin_session import *
from lime_etl.adapter.local_timestamp_adapter import *
from lime_etl.adapter.sqlalchemy_admin_unit_of_work import *
from lime_etl.adapter.sqlalchemy_batch_log_repository import *
from lime_etl.adapter.sqlalchemy_batch_logger import *
from lime_etl.adapter.sqlalchemy_batch_repository import *
from lime_etl.adapter.sqlalchemy_job_log_repository import *
from lime_etl.adapter.sqlalchemy_job_logger import *
from lime_etl.adapter.sqlalchemy_job_repository import *



from lime_etl.domain import exceptions
from lime_etl.domain.admin_unit_of_work import *
from lime_etl.domain.batch_delta import *
from lime_etl.domain.batch_log_repository import *
from lime_etl.domain.batch_logger import *
from lime_etl.domain.batch_repository import *
from lime_etl.domain.batch_spec import *
from lime_etl.domain.value_objects import *
from lime_etl.domain.batch_log_entry import *
from lime_etl.domain.batch_status import *
from lime_etl.domain.job_result import *
from lime_etl.domain.job_dependency_errors import *
from lime_etl.domain.job_log_entry import *
from lime_etl.domain.job_logger import *
from lime_etl.domain.job_log_repository import *
from lime_etl.domain.job_repository import *
from lime_etl.domain.job_spec import *
from lime_etl.domain.job_status import *
from lime_etl.domain.job_test_result import *
from lime_etl.domain.timestamp_adapter import *



import abc

__all__ = ("JobLogger",)


class JobLogger(abc.ABC):
    @abc.abstractmethod
    def log_error(self, message: str, /) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def log_info(self, message: str, /) -> None:
        raise NotImplementedError



import abc

from lime_etl.domain import value_objects, job_logger

__all__ = ("BatchLogger",)


class BatchLogger(abc.ABC):
    @abc.abstractmethod
    def create_job_logger(
        self, /, job_id: value_objects.UniqueId
    ) -> job_logger.JobLogger:
        raise NotImplementedError

    @abc.abstractmethod
    def log_error(self, message: str, /) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def log_info(self, message: str, /) -> None:
        raise NotImplementedError



from __future__ import annotations

import datetime
import typing

from lime_etl import domain

__all__ = ("LocalTimestampAdapter",)


class LocalTimestampAdapter(domain.TimestampAdapter):
    @classmethod
    def interface(cls) -> typing.Type[domain.TimestampAdapter]:
        return domain.TimestampAdapter

    def now(self) -> domain.Timestamp:
        return domain.Timestamp(datetime.datetime.now())



import datetime
import typing

from lime_etl import domain

__all__ = ("DeleteOldLogs",)


class DeleteOldLogs(domain.JobSpec[domain.admin_unit_of_work.AdminUnitOfWork]):
    def __init__(
        self,
        days_logs_to_keep: domain.DaysToKeep = domain.DaysToKeep(3),
        min_seconds_between_runs: domain.MinSecondsBetweenRefreshes = domain.MinSecondsBetweenRefreshes(0),
    ):
        self._days_logs_to_keep = days_logs_to_keep
        self._min_seconds_between_runs = min_seconds_between_runs

    @property
    def job_name(self) -> domain.JobName:
        return domain.JobName("delete_old_logs")

    @property
    def min_seconds_between_refreshes(self) -> domain.MinSecondsBetweenRefreshes:
        return self._min_seconds_between_runs

    def run(
        self,
        admin_uow: domain.AdminUnitOfWork,
        batch_uow: domain.AdminUnitOfWork,
        logger: domain.JobLogger,
    ) -> domain.JobStatus:
        with batch_uow:
            batch_uow.batch_log_repo.delete_old_entries(days_to_keep=self._days_logs_to_keep)
            logger.log_info(
                f"Deleted batch log entries older than {self._days_logs_to_keep.value} days old."
            )

            batch_uow.job_log_repo.delete_old_entries(days_to_keep=self._days_logs_to_keep)
            logger.log_info(
                f"Deleted job log entries older than {self._days_logs_to_keep.value} days old."
            )

            batch_uow.batch_repo.delete_old_entries(self._days_logs_to_keep)
            logger.log_info(
                f"Deleted batch results older than {self._days_logs_to_keep.value} days old."
            )
            batch_uow.save()

        return domain.JobStatus.success()

    def test(
        self,
        admin_uow: domain.AdminUnitOfWork,
        batch_uow: domain.AdminUnitOfWork,
        logger: domain.JobLogger,
    ) -> typing.List[domain.SimpleJobTestResult]:
        with batch_uow:
            now = batch_uow.ts_adapter.now().value
            cutoff_date = datetime.datetime.combine(
                (now - datetime.timedelta(days=self._days_logs_to_keep.value)).date(),
                datetime.datetime.min.time(),
            )
            earliest_ts = batch_uow.batch_log_repo.get_earliest_timestamp()

        if earliest_ts and earliest_ts < cutoff_date:
            return [
                domain.SimpleJobTestResult(
                    test_name=domain.TestName("No log entries more than 3 days old"),
                    test_success_or_failure=domain.Result.failure(
                        f"The earliest batch log entry is from "
                        f"{earliest_ts.strftime('%Y-%m-%d %H:%M:%S')}"
                    ),
                )
            ]
        else:
            return [
                domain.SimpleJobTestResult(
                    test_name=domain.TestName("No log entries more than 3 days old"),
                    test_success_or_failure=domain.Result.success(),
                )
            ]



import abc
import typing

import lime_uow as lu

from lime_etl.domain import job_result, job_status, job_test_result, value_objects


__all__ = ("JobRepository",)


class JobRepository(lu.Repository[job_result.JobResultDTO], abc.ABC):
    @abc.abstractmethod
    def last_job_run_status(
        self, /, job_name: value_objects.JobName
    ) -> typing.Optional[job_status.JobStatus]:
        raise NotImplementedError

    @abc.abstractmethod
    def latest_test_results(
        self, /, job_name: value_objects.JobName,
    ) -> typing.FrozenSet[job_test_result.JobTestResult]:
        raise NotImplementedError

    @abc.abstractmethod
    def get_last_successful_ts(
        self, /, job_name: value_objects.JobName
    ) -> typing.Optional[value_objects.Timestamp]:
        raise NotImplementedError



from __future__ import annotations

import datetime
import dataclasses

from lime_etl.domain import value_objects


__all__ = (
    "JobLogEntryDTO",
    "JobLogEntry",
)


@dataclasses.dataclass(unsafe_hash=True)
class JobLogEntryDTO:
    id: str
    batch_id: str
    job_id: str
    log_level: value_objects.LogLevelOption
    message: str
    ts: datetime.datetime

    def to_domain(self) -> JobLogEntry:
        return JobLogEntry(
            id=value_objects.UniqueId(self.id),
            batch_id=value_objects.UniqueId(self.batch_id),
            job_id=value_objects.UniqueId(self.job_id),
            log_level=value_objects.LogLevel(self.log_level),
            message=value_objects.LogMessage(self.message),
            ts=value_objects.Timestamp(self.ts),
        )


@dataclasses.dataclass(frozen=True)
class JobLogEntry:
    id: value_objects.UniqueId
    batch_id: value_objects.UniqueId
    job_id: value_objects.UniqueId
    log_level: value_objects.LogLevel
    message: value_objects.LogMessage
    ts: value_objects.Timestamp

    def to_dto(self) -> JobLogEntryDTO:
        return JobLogEntryDTO(
            id=self.id.value,
            batch_id=self.batch_id.value,
            job_id=self.job_id.value,
            log_level=self.log_level.value,
            message=self.message.value,
            ts=self.ts.value,
        )

    def __str__(self) -> str:
        ts_str = self.ts.value.strftime("%H:%M:%S")
        return f"{ts_str} - {self.log_level} - {self.message.value}"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['lime_etl',
 'lime_etl.adapter',
 'lime_etl.domain',
 'lime_etl.service',
 'lime_etl.service.admin']

package_data = \
{'': ['*']}

install_requires = \
['lime-uow>=0.1.44,<0.2.0', 'sqlalchemy>=1.3.19,<1.4']

setup_kwargs = {
    'name': 'lime-etl',
    'version': '0.1.28',
    'description': 'Simple ETL job runner',
    'long_description': None,
    'author': 'Mark Stefanovic',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import typing

from lime_uow import sqlalchemy_resources as lsa
from sqlalchemy import orm
from sqlalchemy.sql.expression import desc

from lime_etl import domain

__all__ = ("SqlAlchemyJobRepository",)

from lime_etl.domain import job_status, value_objects


class SqlAlchemyJobRepository(
    domain.JobRepository, lsa.SqlAlchemyRepository[domain.JobResultDTO]
):
    def __init__(
        self,
        session: orm.Session,
        ts_adapter: domain.TimestampAdapter,
    ):
        self._ts_adapter = ts_adapter
        super().__init__(session)

    def last_job_run_status(
        self, /, job_name: value_objects.JobName,
    ) -> typing.Optional[job_status.JobStatus]:
        # noinspection PyUnresolvedReferences,PyTypeChecker
        jr: typing.Optional[domain.JobResultDTO] = (
            self._session.query(domain.JobResultDTO)
            .filter(domain.JobResultDTO.job_name.ilike(job_name.value))  # type: ignore
            .filter(domain.JobResultDTO.skipped.is_(False))  # type: ignore
            .order_by(desc(domain.JobResultDTO.ts))  # type: ignore
            .offset(1)
            .first()
        )
        if jr is None:
            return None
        else:
            return jr.to_domain().status

    def latest_test_results(
        self, /, job_name: value_objects.JobName,
    ) -> typing.FrozenSet[domain.JobTestResult]:
        # noinspection PyUnresolvedReferences,PyTypeChecker
        jr: typing.Optional[domain.JobResultDTO] = (
            self._session.query(domain.JobResultDTO)
            .filter(domain.JobResultDTO.job_name.ilike(job_name.value))  # type: ignore
            .filter(domain.JobResultDTO.skipped.is_(False))  # type: ignore
            .order_by(desc(domain.JobResultDTO.ts))  # type: ignore
            .offset(1)
            .first()
        )
        if jr is None:
            return frozenset()
        else:
            return jr.to_domain().test_results

    @property
    def entity_type(self) -> typing.Type[domain.JobResultDTO]:
        return domain.JobResultDTO

    def get_last_successful_ts(
        self, job_name: domain.JobName, /
    ) -> typing.Optional[domain.Timestamp]:
        # noinspection PyUnresolvedReferences,PyTypeChecker
        jr: typing.Optional[domain.JobResultDTO] = (
            self._session.query(domain.JobResultDTO)
            .filter(domain.JobResultDTO.job_name.ilike(job_name.value))  # type: ignore
            .filter(domain.JobResultDTO.execution_error_occurred.is_(False))  # type: ignore
            .filter(domain.JobResultDTO.skipped.is_(False))  # type: ignore
            .order_by(desc(domain.JobResultDTO.ts))  # type: ignore
            .first()
        )
        if jr is None:
            return None
        else:
            return domain.Timestamp(jr.ts)

    @classmethod
    def interface(cls) -> typing.Type[domain.JobRepository]:
        return domain.JobRepository



from __future__ import annotations

import typing

from lime_uow import sqlalchemy_resources as lsa
from sqlalchemy import orm

__all__ = ("SqlAlchemyAdminSession",)


class SqlAlchemyAdminSession(lsa.SqlAlchemySession):
    def __init__(self, session_factory: orm.sessionmaker):
        super().__init__(session_factory)

    @classmethod
    def interface(cls) -> typing.Type[SqlAlchemyAdminSession]:
        return SqlAlchemyAdminSession



from __future__ import annotations

import dataclasses

from lime_etl.domain import value_objects


__all__ = (
    "JobRanSuccessfully",
    "JobFailed",
    "JobInProgress",
    "JobStatus",
    "JobSkipped",
)


@dataclasses.dataclass(frozen=True)
class JobStatus:
    @staticmethod
    def failed(error_message: str, /) -> JobFailed:
        msg = value_objects.LogMessage(error_message)
        return JobFailed(msg)

    @staticmethod
    def in_progress() -> JobInProgress:
        return JobInProgress()

    @staticmethod
    def skipped(reason: str, /) -> JobSkipped:
        msg = value_objects.LogMessage(reason)
        return JobSkipped(reason=msg)

    @staticmethod
    def success() -> JobRanSuccessfully:
        return JobRanSuccessfully()


@dataclasses.dataclass(frozen=True)
class JobFailed(JobStatus):
    error_message: value_objects.LogMessage


@dataclasses.dataclass(frozen=True)
class JobRanSuccessfully(JobStatus):
    ...


@dataclasses.dataclass(frozen=True)
class JobSkipped(JobStatus):
    reason: value_objects.LogMessage


@dataclasses.dataclass(frozen=True)
class JobInProgress(JobStatus):
    ...



import abc

import lime_uow as lu

from lime_etl.domain import job_log_entry, value_objects

__all__ = (
    "JobLogRepository",
)


class JobLogRepository(
    lu.Repository[job_log_entry.JobLogEntryDTO],
    abc.ABC,
):
    @abc.abstractmethod
    def delete_old_entries(self, days_to_keep: value_objects.DaysToKeep) -> int:
        raise NotImplementedError





import datetime
import typing

import sqlalchemy as sa
from lime_uow import sqlalchemy_resources as lsa
from sqlalchemy.orm.session import Session

from lime_etl import domain

__all__ = ("SqlAlchemyBatchRepository",)


class SqlAlchemyBatchRepository(
    domain.BatchRepository, lsa.SqlAlchemyRepository[domain.BatchStatusDTO]
):
    def __init__(
        self,
        session: Session,
        ts_adapter: domain.TimestampAdapter,
    ):
        super().__init__(session)
        self._ts_adapter = ts_adapter

    def delete_old_entries(self, days_to_keep: domain.DaysToKeep, /) -> int:
        ts = self._ts_adapter.now().value
        cutoff: datetime.datetime = ts - datetime.timedelta(days=days_to_keep.value)
        # We need to delete batches one by one to trigger cascade deletes, a bulk update will
        # not trigger them, and we don't want to rely on specific database implementations, so
        # we cannot use ondelete='CASCADE' on the foreign key columns.
        batches: typing.List[domain.BatchStatusDTO] = (
            self.session.query(domain.BatchStatusDTO)
            .filter(domain.BatchStatusDTO.ts < cutoff)
            .all()
        )
        for b in batches:
            self.session.delete(b)
        return len(batches)

    @property
    def entity_type(self) -> typing.Type[domain.BatchStatusDTO]:
        return domain.BatchStatusDTO

    def get_latest(
        self, /, batch_name: domain.BatchName
    ) -> typing.Optional[domain.BatchStatusDTO]:
        # noinspection PyTypeChecker
        return (
            self.session.query(domain.BatchStatusDTO)
            .filter(domain.BatchStatusDTO.name == batch_name.value)
            .order_by(sa.desc(domain.BatchStatusDTO.ts))  # type: ignore
            .first()
        )

    def get_latest_batch_delta(
        self, /, batch_name: domain.BatchName
    ) -> domain.BatchDelta:
        if (current_dto := self.get_latest(batch_name)) is None:
            raise domain.exceptions.BatchNotFound(
                f"No previous results for batch {batch_name.value!r} were found."
            )

        if (previous_dto := self.get_previous(batch_name)) is None:
            previous = None
        else:
            previous = previous_dto.to_domain()

        return domain.BatchDelta(
            current_results=current_dto.to_domain(),
            previous_results=previous,
        )

    def get_previous(
        self, /, batch_name: domain.BatchName
    ) -> typing.Optional[domain.BatchStatusDTO]:
        # noinspection PyTypeChecker
        return (
            self.session.query(domain.BatchStatusDTO)
            .filter(domain.BatchStatusDTO.name == batch_name.value)
            .order_by(sa.desc(domain.BatchStatusDTO.ts))  # type: ignore
            .offset(1)
            .first()
        )

    @classmethod
    def interface(cls) -> typing.Type[domain.BatchRepository]:
        return domain.BatchRepository



import abc
import functools
import typing

import lime_uow as lu

from lime_etl.domain import job_spec, value_objects

UoW = typing.TypeVar("UoW", bound=lu.UnitOfWork)

__all__ = ("BatchSpec",)


class BatchSpec(abc.ABC, typing.Generic[UoW]):
    @functools.cached_property
    def batch_id(self) -> value_objects.UniqueId:
        return value_objects.UniqueId.generate()

    @property
    @abc.abstractmethod
    def batch_name(self) -> value_objects.BatchName:
        raise NotImplementedError

    @abc.abstractmethod
    def create_jobs(self, uow: UoW) -> typing.List[job_spec.JobSpec[UoW]]:
        raise NotImplementedError

    @abc.abstractmethod
    def create_uow(self) -> UoW:
        raise NotImplementedError

    @property
    def skip_tests(self) -> value_objects.Flag:
        return value_objects.Flag(False)

    @property
    def timeout_seconds(self) -> value_objects.TimeoutSeconds:
        return value_objects.TimeoutSeconds(None)

    def __repr__(self) -> str:
        return f"<BatchSpec: {self.__class__.__name__}>: {self.batch_name.value}"

    def __hash__(self) -> int:
        return hash(self.batch_name.value)

    def __eq__(self, other: object) -> bool:
        if other.__class__ is self.__class__:
            return (
                self.batch_name.value
                == typing.cast(BatchSpec[typing.Any], other).batch_name.value
            )
        else:
            return NotImplemented



from __future__ import annotations

import dataclasses
import datetime
import typing

from lime_etl.domain import value_objects

__all__ = (
    "JobTestResultDTO",
    "JobTestResult",
    "SimpleJobTestResult",
)


@dataclasses.dataclass(unsafe_hash=True)
class JobTestResultDTO:
    id: str
    job_id: str
    test_name: str
    test_passed: bool
    test_failure_message: typing.Optional[str]
    execution_millis: int
    execution_error_occurred: bool
    execution_error_message: typing.Optional[str]
    ts: datetime.datetime

    def to_domain(self) -> JobTestResult:
        if self.test_passed:
            test_success_or_failure = value_objects.Result.success()
        else:
            test_success_or_failure = value_objects.Result.failure(
                self.test_failure_message or "No error message was provided."
            )

        if self.execution_error_occurred:
            execution_success_or_failure = value_objects.Result.failure(
                self.execution_error_message or "No error message was provided."
            )
        else:
            execution_success_or_failure = value_objects.Result.success()

        return JobTestResult(
            id=value_objects.UniqueId(self.id),
            job_id=value_objects.UniqueId(self.job_id),
            test_name=value_objects.TestName(self.test_name),
            test_success_or_failure=test_success_or_failure,
            execution_millis=value_objects.ExecutionMillis(self.execution_millis),
            execution_success_or_failure=execution_success_or_failure,
            ts=value_objects.Timestamp(self.ts),
        )


@dataclasses.dataclass(frozen=True)
class JobTestResult:
    id: value_objects.UniqueId
    job_id: value_objects.UniqueId
    test_name: value_objects.TestName
    test_success_or_failure: value_objects.Result
    execution_millis: value_objects.ExecutionMillis
    execution_success_or_failure: value_objects.Result
    ts: value_objects.Timestamp

    @property
    def test_failed(self) -> bool:
        return self.test_success_or_failure.is_failure

    @property
    def test_passed(self) -> bool:
        return not self.test_failed

    def to_dto(self) -> JobTestResultDTO:
        return JobTestResultDTO(
            id=self.id.value,
            job_id=self.job_id.value,
            test_name=self.test_name.value,
            test_failure_message=self.test_success_or_failure.failure_message_or_none,
            test_passed=self.test_passed,
            execution_millis=self.execution_millis.value,
            execution_error_occurred=self.execution_success_or_failure.is_failure,
            execution_error_message=self.execution_success_or_failure.failure_message_or_none,
            ts=self.ts.value,
        )


@dataclasses.dataclass(frozen=True)
class SimpleJobTestResult:
    test_name: value_objects.TestName
    test_success_or_failure: value_objects.Result

    @property
    def test_failed(self) -> bool:
        return self.test_success_or_failure.is_failure

    @property
    def test_passed(self) -> bool:
        return self.test_success_or_failure.is_success



import dataclasses
import typing

from lime_etl.domain import value_objects

__all__ = ("JobDependencyErrors",)


@dataclasses.dataclass(frozen=True, order=True)
class JobDependencyErrors:
    job_name: value_objects.JobName
    missing_dependencies: typing.FrozenSet[value_objects.JobName]
    jobs_out_of_order: typing.FrozenSet[value_objects.JobName]

    def __str__(self) -> str:
        if self.missing_dependencies:
            missing_deps_str: typing.Optional[str] = (
                f" has the following unresolved dependencies: "
                + f"{', '.join(sorted('[' + dep.value + ']' for dep in self.missing_dependencies))}"
            )
        else:
            missing_deps_str = None

        if self.jobs_out_of_order:
            jobs_out_of_order_str: typing.Optional[str] = (
                f" depends on the following jobs which come after it: "
                + f"{', '.join(sorted('[' + dep.value + ']'  for dep in self.jobs_out_of_order))}"
            )
        else:
            jobs_out_of_order_str = None

        if missing_deps_str and jobs_out_of_order_str:
            return f"[{self.job_name.value}]{missing_deps_str}.  It also {jobs_out_of_order_str}."
        elif missing_deps_str:
            return f"[{self.job_name.value}]{missing_deps_str}."
        elif jobs_out_of_order_str:
            return f"[{self.job_name.value}]{jobs_out_of_order_str}."
        else:
            raise ValueError(
                "There are no missing dependencies or jobs out of order, so there is nothing to render."
            )
